@@ -1,70 +1,49 @@
 //! This module contains a [InteractOptions] which allows a castomization of
 //! [crate::Session::interact] flow.
 
-use crate::{session::Session, ControlCode, Error};
+use crate::{
+    process::{Healthcheck, NonBlocking},
+    session::Proc,
+    session::Session,
+    stream::stdin::Stdin,
+    Captures, ControlCode, Error,
+};
 use std::{
     borrow::Cow,
     collections::HashMap,
-    io::{self, Write},
+    io::{self, Read, Stdout, Write},
 };
 
-#[cfg(unix)]
-use crate::{stream::Stream, WaitStatus};
-#[cfg(unix)]
-use nix::{
-    libc::STDIN_FILENO,
-    sys::termios,
-    unistd::{dup, isatty},
-};
-#[cfg(unix)]
-use ptyprocess::set_raw;
-#[cfg(unix)]
-use std::os::unix::prelude::FromRawFd;
-
-#[cfg(not(feature = "async"))]
-use std::io::Read;
-
-#[cfg(windows)]
-use conpty::console::Console;
-
-/// InteractOptions represents options of an interact session.
-pub struct InteractOptions<R, W, C = ()> {
-    input: R,
-    output: W,
-    input_from: InputFrom,
+/// InteractOptions represents options of an interactive session.
+pub struct InteractOptions<P, S, R, W, C> {
     escape_character: u8,
     input_filter: Option<FilterFn>,
     output_filter: Option<FilterFn>,
-    input_handlers: HashMap<String, ActionFn<R, W, C>>,
+    input_handlers: HashMap<String, ActionFn<Session<P, S>, R, W, C>>,
     #[allow(clippy::type_complexity)]
-    output_handlers: Vec<(Box<dyn crate::Needle>, OutputFn<R, W, C>)>,
-    idle_handler: Option<ActionFn<R, W, C>>,
+    output_handlers: Vec<(Box<dyn crate::Needle>, OutputFn<Session<P, S>, R, W, C>)>,
+    idle_handler: Option<ActionFn<Session<P, S>, R, W, C>>,
     state: C,
 }
 
-enum InputFrom {
-    Terminal,
-    Other,
-}
+type ActionFn<S, R, W, C> = Box<dyn FnMut(Context<'_, S, R, W, C>) -> Result<(), Error>>;
 
-type ActionFn<R, W, C> = Box<dyn FnMut(Context<'_, R, W, C>) -> Result<(), Error>>;
-
-type OutputFn<R, W, C> = Box<dyn FnMut(Context<'_, R, W, C>, crate::Found) -> Result<(), Error>>;
+type OutputFn<S, R, W, C> = Box<dyn FnMut(Context<'_, S, R, W, C>, Captures) -> Result<(), Error>>;
 
 type FilterFn = Box<dyn FnMut(&[u8]) -> Result<Cow<[u8]>, Error>>;
 
 /// Context provides an interface to use a [Session], IO streams
 /// and a state.
-pub struct Context<'a, R, W, C> {
-    session: &'a mut Session,
+pub struct Context<'a, S, R, W, C> {
+    session: &'a mut S,
     input: &'a mut R,
     output: &'a mut W,
     state: &'a mut C,
 }
 
-impl<'a, R, W, C> Context<'a, R, W, C> {
+impl<S, R, W, C> Context<'_, S, R, W, C> {
     /// Get a reference to the context's session.
-    pub fn session(&mut self) -> &mut Session {
+    pub fn session(&mut self) -> &mut S {
         self.session
     }
 
@@ -84,38 +63,9 @@ impl<'a, R, W, C> Context<'a, R, W, C> {
     }
 }
 
-impl InteractOptions<NonBlockingStdin, io::Stdout> {
-    /// Constructs a interact options to interact via STDIN.
-    ///
-    /// Usage [InteractOptions::streamed] directly with [std::io::stdin],
-    /// most likely will provide a correct interact processing.
-    /// It depends on terminal settings.
-    pub fn terminal() -> Result<Self, Error> {
-        let input = NonBlockingStdin::new()?;
-        Ok(Self {
-            input,
-            output: io::stdout(),
-            input_from: InputFrom::Terminal,
-            escape_character: Self::default_escape_char(),
-            input_filter: None,
-            output_filter: None,
-            idle_handler: None,
-            input_handlers: HashMap::new(),
-            output_handlers: Vec::new(),
-            state: (),
-        })
-    }
-}
-
-impl<R, W> InteractOptions<R, W> {
-    /// Create interact options with custom input and output streams.
-    ///
-    /// To construct default terminal session see [InteractOptions::terminal]
-    pub fn streamed(input: R, output: W) -> Result<Self, Error> {
-        Ok(Self {
-            input,
-            output,
-            input_from: InputFrom::Other,
+impl<P, S, R, W> Default for InteractOptions<P, S, R, W, ()> {
+    fn default() -> Self {
+        Self {
             escape_character: Self::default_escape_char(),
             idle_handler: None,
             input_handlers: HashMap::new(),
@@ -123,25 +73,22 @@ impl<R, W> InteractOptions<R, W> {
             input_filter: None,
             output_filter: None,
             state: (),
-        })
+        }
     }
 }
 
-impl<R, W, C> InteractOptions<R, W, C> {
+impl<P, S, R, W, C> InteractOptions<P, S, R, W, C> {
     /// State sets state which will be available in callback calls, throught context variable.
     ///
     /// Please beware that it cleans already set list of callbacks.
     /// So you need to call this method BEFORE you specify callbacks.
     ///
     /// Default state type is a unit type `()`.
-    pub fn state<C1>(self, state: C1) -> InteractOptions<R, W, C1> {
+    pub fn state<C1>(self, state: C1) -> InteractOptions<P, S, R, W, C1> {
         InteractOptions {
             state,
             escape_character: self.escape_character,
-            input: self.input,
             input_filter: self.input_filter,
-            input_from: self.input_from,
-            output: self.output,
             output_filter: self.output_filter,
             idle_handler: None,
             input_handlers: HashMap::new(),
@@ -160,7 +107,7 @@ impl<R, W, C> InteractOptions<R, W, C> {
     }
 }
 
-impl<R, W, C> InteractOptions<R, W, C> {
+impl<P, S, R, W, C> InteractOptions<P, S, R, W, C> {
     /// Sets an escape character after seen which the interact interactions will be stopped
     /// and controll will be returned to a caller process.
     pub fn escape_character(mut self, c: u8) -> Self {
@@ -201,9 +148,9 @@ impl<R, W, C> InteractOptions<R, W, C> {
     /// See <https://github.com/zhiburt/expectrl/issues/16>.
     pub fn on_input<F>(mut self, input: impl Into<String>, f: F) -> Self
     where
-        F: FnMut(Context<'_, R, W, C>) -> Result<(), Error> + 'static,
+        F: FnMut(Context<'_, Session<P, S>, R, W, C>) -> Result<(), Error> + 'static,
     {
-        self.input_handlers.insert(input.into(), Box::new(f));
+        let _ = self.input_handlers.insert(input.into(), Box::new(f));
         self
     }
 
@@ -215,7 +162,7 @@ impl<R, W, C> InteractOptions<R, W, C> {
     pub fn on_output<N, F>(mut self, needle: N, f: F) -> Self
     where
         N: crate::Needle + 'static,
-        F: FnMut(Context<'_, R, W, C>, crate::Found) -> Result<(), Error> + 'static,
+        F: FnMut(Context<'_, Session<P, S>, R, W, C>, Captures) -> Result<(), Error> + 'static,
     {
         self.output_handlers.push((Box::new(needle), Box::new(f)));
         self
@@ -224,7 +171,7 @@ impl<R, W, C> InteractOptions<R, W, C> {
     /// Puts a handler which will be called on each interaction.
     pub fn on_idle<F>(mut self, f: F) -> Self
     where
-        F: FnMut(Context<'_, R, W, C>) -> Result<(), Error> + 'static,
+        F: FnMut(Context<'_, Session<P, S>, R, W, C>) -> Result<(), Error> + 'static,
     {
         self.idle_handler = Some(Box::new(f));
         self
@@ -234,7 +181,13 @@ impl<R, W, C> InteractOptions<R, W, C> {
         ControlCode::GroupSeparator.into() // Ctrl-]
     }
 
-    fn check_input(&mut self, session: &mut Session, bytes: &[u8]) -> Result<Match, Error> {
+    fn check_input(
+        &mut self,
+        input: &mut R,
+        output: &mut W,
+        session: &mut Session<P, S>,
+        bytes: &[u8],
+    ) -> Result<Match, Error> {
         for (pattern, callback) in self.input_handlers.iter_mut() {
             if !pattern.is_empty() && !bytes.is_empty() {
                 match contains_in_bytes(bytes, pattern.as_bytes()) {
@@ -244,10 +197,10 @@ impl<R, W, C> InteractOptions<R, W, C> {
                     }
                     Match::Yes(n) => {
                         let context = Context {
-                            input: &mut self.input,
-                            output: &mut self.output,
                             state: &mut self.state,
                             session,
+                            input,
+                            output,
                         };
                         (callback)(context)?;
                         return Ok(Match::Yes(n));
@@ -261,7 +214,9 @@ impl<R, W, C> InteractOptions<R, W, C> {
 
     fn check_output(
         &mut self,
-        session: &mut Session,
+        input: &mut R,
+        output: &mut W,
+        session: &mut Session<P, S>,
         buf: &mut Vec<u8>,
         eof: bool,
     ) -> Result<(), Error> {
@@ -269,16 +224,16 @@ impl<R, W, C> InteractOptions<R, W, C> {
             for (search, callback) in self.output_handlers.iter_mut() {
                 let found = search.check(buf, eof)?;
                 if !found.is_empty() {
-                    let end_index = crate::Found::right_most_index(&found);
+                    let end_index = Captures::right_most_index(&found);
                     let involved_bytes = buf[..end_index].to_vec();
-                    let found = crate::Found::new(involved_bytes, found);
-                    buf.drain(..end_index);
+                    let found = Captures::new(involved_bytes, found);
+                    let _ = buf.drain(..end_index);
 
                     let context = Context {
-                        input: &mut self.input,
-                        output: &mut self.output,
                         state: &mut self.state,
                         session,
+                        input,
+                        output,
                     };
                     (callback)(context, found)?;
 
@@ -290,12 +245,17 @@ impl<R, W, C> InteractOptions<R, W, C> {
         }
     }
 
-    fn call_idle_handler(&mut self, session: &mut Session) -> Result<(), Error> {
+    fn call_idle_handler(
+        &mut self,
+        input: &mut R,
+        output: &mut W,
+        session: &mut Session<P, S>,
+    ) -> Result<(), Error> {
         let context = Context {
-            input: &mut self.input,
-            output: &mut self.output,
             state: &mut self.state,
             session,
+            input,
+            output,
         };
         if let Some(callback) = self.idle_handler.as_mut() {
             (callback)(context)?;
@@ -306,130 +266,128 @@ impl<R, W, C> InteractOptions<R, W, C> {
 }
 
 #[cfg(not(feature = "async"))]
-impl<R, W, C> InteractOptions<R, W, C>
+impl<P, S, R, W, C> InteractOptions<P, S, R, W, C>
 where
+    P: Healthcheck,
+    S: NonBlocking + Read + Write,
     R: Read,
     W: Write,
 {
     /// Runs interact interactively.
     /// See [Session::interact]
-    #[cfg(unix)]
-    pub fn interact(&mut self, session: &mut Session) -> Result<WaitStatus, Error> {
-        match self.input_from {
-            InputFrom::Terminal => interact_in_terminal(session, self),
-            InputFrom::Other => interact(session, self),
-        }
+    ///
+    /// On process exit it tries to read available bytes from output in order to run callbacks.
+    /// But it is not guaranteed that all output will be read therefore some callbacks might be not called.
+    ///
+    /// To mitigate such an issue you could use [Session::is_empty] to verify that there is nothing in processes output.
+    /// (at the point of the call)
+    pub fn interact(
+        &mut self,
+        session: &mut Session<P, S>,
+        mut input: R,
+        mut output: W,
+    ) -> Result<(), Error> {
+        interact(self, session, &mut input, &mut output)
     }
+}
 
-    /// Runs interact interactively.
-    /// See [Session::interact]
-    #[cfg(windows)]
-    pub fn interact(self, session: &mut Session) -> Result<(), Error> {
-        match self.input_from {
-            InputFrom::Terminal => interact_in_terminal(session, self),
-            InputFrom::Other => interact(session, self),
-        }
+#[cfg(not(feature = "async"))]
+impl<S, C> InteractOptions<Proc, S, Stdin, Stdout, C>
+where
+    S: NonBlocking + Read + Write,
+{
+    /// This function runs an interact loop with a terminal, using the provided settings.
+    ///
+    /// If you don't use any settings you can use [Session::interact].
+    pub fn interact_in_terminal(&mut self, session: &mut Session<Proc, S>) -> Result<(), Error> {
+        let mut stdin = crate::stream::stdin::Stdin::new(session)?;
+        let r = interact(self, session, &mut stdin, &mut std::io::stdout());
+        stdin.close(session)?;
+        r
     }
 }
 
 #[cfg(feature = "async")]
-impl<R, W> InteractOptions<R, W>
+impl<P, S, R, W, C> InteractOptions<P, S, R, W, C>
 where
-    R: futures_lite::AsyncRead + std::marker::Unpin,
+    P: Healthcheck + Unpin,
+    S: futures_lite::AsyncRead + futures_lite::AsyncWrite + Unpin,
+    R: futures_lite::AsyncRead + Unpin,
     W: Write,
 {
     /// Runs interact interactively.
     /// See [Session::interact]
-    pub async fn interact(self, session: &mut Session) -> Result<WaitStatus, Error> {
-        match self.input_from {
-            InputFrom::Terminal => interact_in_terminal(session, self).await,
-            InputFrom::Other => interact(session, self).await,
-        }
+    pub async fn interact(
+        &mut self,
+        session: &mut Session<P, S>,
+        mut input: R,
+        mut output: W,
+    ) -> Result<(), Error> {
+        interact(self, session, &mut input, &mut output).await
     }
 }
 
-#[cfg(all(unix, not(feature = "async")))]
-fn interact_in_terminal<R, W, C>(
-    session: &mut Session,
-    options: &mut InteractOptions<R, W, C>,
-) -> Result<WaitStatus, Error>
+#[cfg(feature = "async")]
+impl<S, C> InteractOptions<Proc, S, Stdin, Stdout, C>
 where
-    R: Read,
-    W: Write,
+    S: futures_lite::AsyncRead + futures_lite::AsyncWrite + Unpin,
 {
-    // flush buffers
-    session.flush()?;
-
-    let origin_pty_echo = session.get_echo()?;
-    // tcgetattr issues error if a provided fd is not a tty,
-    // but we can work with such input as it may be redirected.
-    let origin_stdin_flags = termios::tcgetattr(STDIN_FILENO);
-
-    // verify: possible controlling fd can be stdout and stderr as well?
-    // https://stackoverflow.com/questions/35873843/when-setting-terminal-attributes-via-tcsetattrfd-can-fd-be-either-stdout
-    let isatty_terminal = isatty(STDIN_FILENO)?;
-
-    if isatty_terminal {
-        set_raw(STDIN_FILENO)?;
+    /// This function runs an interact loop with a terminal, using the provided settings.
+    ///
+    /// If you don't use any settings you can use [Session::interact].
+    pub async fn interact_in_terminal(
+        &mut self,
+        session: &mut Session<Proc, S>,
+    ) -> Result<(), Error> {
+        let mut stdin = crate::stream::stdin::Stdin::new(session)?;
+        let r = interact(self, session, &mut stdin, &mut std::io::stdout()).await;
+        stdin.close(session)?;
+        r
     }
-
-    session.set_echo(true)?;
-
-    let result = interact(session, options);
-
-    if isatty_terminal {
-        // it's suppose to be always OK.
-        // but we don't use unwrap just in case.
-        let origin_stdin_flags = origin_stdin_flags?;
-
-        termios::tcsetattr(
-            STDIN_FILENO,
-            termios::SetArg::TCSAFLUSH,
-            &origin_stdin_flags,
-        )?;
-    }
-
-    session.set_echo(origin_pty_echo)?;
-
-    result
 }
 
-#[cfg(unix)]
 #[cfg(not(feature = "async"))]
-fn interact<R, W, C>(
-    session: &mut Session,
-    options: &mut InteractOptions<R, W, C>,
-) -> Result<WaitStatus, Error>
+fn interact<P, S, R, W, C>(
+    options: &mut InteractOptions<P, S, R, W, C>,
+    session: &mut Session<P, S>,
+    input: &mut R,
+    output: &mut W,
+) -> Result<(), Error>
 where
+    P: Healthcheck,
+    S: NonBlocking + Read + Write,
     R: Read,
     W: Write,
 {
+    let mut output_buffer = Vec::new();
     let options_has_input_checks = !options.input_handlers.is_empty();
     let mut input_buffer = if options_has_input_checks {
         Some(Vec::new())
     } else {
         None
     };
-
-    let mut output_buffer = Vec::new();
+    let mut exited = false;
 
     let mut buf = [0; 512];
     loop {
-        let status = session.status()?;
-        if !matches!(status, WaitStatus::StillAlive) {
-            return Ok(status);
+        // In case where proceses exits we are trying to
+        // fill buffer to run callbacks if there was something in.
+        //
+        // We ignore errors because there might be errors like EOCHILD etc.
+        let status = session.is_alive();
+        if matches!(status, Ok(false)) {
+            exited = true;
         }
 
         match session.try_read(&mut buf) {
             Ok(n) => {
                 let eof = n == 0;
+                if eof {
+                    exited = true;
+                }
 
                 output_buffer.extend_from_slice(&buf[..n]);
-                options.check_output(session, &mut output_buffer, eof)?;
-
-                if n == 0 {
-                    return Ok(status);
-                }
+                options.check_output(input, output, session, &mut output_buffer, eof)?;
 
                 let bytes = if let Some(filter) = options.output_filter.as_mut() {
                     (filter)(&buf[..n])?
@@ -437,302 +395,22 @@ where
                     Cow::Borrowed(&buf[..n])
                 };
 
-                options.output.write_all(&bytes)?;
-                options.output.flush()?;
+                output.write_all(&bytes)?;
+                output.flush()?;
             }
             Err(err) if err.kind() == io::ErrorKind::WouldBlock => {}
             Err(err) => return Err(err.into()),
         }
 
+        if exited {
+            return Ok(());
+        }
+
         // We dont't print user input back to the screen.
         // In terminal mode it will be ECHOed back automatically.
         // This way we preserve terminal seetings for example when user inputs password.
         // The terminal must have been prepared before.
-        match options.input.read(&mut buf) {
-            Ok(0) => {
-                return Ok(status);
-            }
-            Ok(n) => {
-                let bytes = &buf[..n];
-                let bytes = if let Some(filter) = options.input_filter.as_mut() {
-                    (filter)(bytes)?
-                } else {
-                    Cow::Borrowed(bytes)
-                };
-
-                let buffer = if let Some(check_buffer) = input_buffer.as_mut() {
-                    check_buffer.extend_from_slice(&bytes);
-                    loop {
-                        match options.check_input(session, check_buffer)? {
-                            Match::Yes(n) => {
-                                check_buffer.drain(..n);
-                                if check_buffer.is_empty() {
-                                    break vec![];
-                                }
-                            }
-                            Match::No => {
-                                let buffer = check_buffer.to_vec();
-                                check_buffer.clear();
-                                break buffer;
-                            }
-                            Match::MaybeLater => break vec![],
-                        }
-                    }
-                } else {
-                    bytes.to_vec()
-                };
-
-                let escape_char_position =
-                    buffer.iter().position(|c| *c == options.escape_character);
-                match escape_char_position {
-                    Some(pos) => {
-                        session.write_all(&buffer[..pos])?;
-                        return Ok(status);
-                    }
-                    None => {
-                        session.write_all(&buffer[..])?;
-                    }
-                }
-            }
-            Err(err) if err.kind() == io::ErrorKind::WouldBlock => {}
-            Err(err) => return Err(err.into()),
-        }
-
-        options.call_idle_handler(session)?;
-    }
-}
-
-// copy paste of sync version with async await syntax
-#[cfg(all(unix, feature = "async"))]
-async fn interact_in_terminal<R, W>(
-    session: &mut Session,
-    options: InteractOptions<R, W>,
-) -> Result<WaitStatus, Error>
-where
-    R: futures_lite::AsyncRead + std::marker::Unpin,
-    W: Write,
-{
-    use futures_lite::AsyncWriteExt;
-
-    // flush buffers
-    session.flush().await?;
-
-    let origin_pty_echo = session.get_echo()?;
-    // tcgetattr issues error if a provided fd is not a tty,
-    // but we can work with such input as it may be redirected.
-    let origin_stdin_flags = termios::tcgetattr(STDIN_FILENO);
-
-    // verify: possible controlling fd can be stdout and stderr as well?
-    // https://stackoverflow.com/questions/35873843/when-setting-terminal-attributes-via-tcsetattrfd-can-fd-be-either-stdout
-    let isatty_terminal = isatty(STDIN_FILENO)?;
-
-    if isatty_terminal {
-        set_raw(STDIN_FILENO)?;
-    }
-
-    session.set_echo(true)?;
-
-    let result = interact(session, options).await;
-
-    if isatty_terminal {
-        // it's suppose to be always OK.
-        // but we don't use unwrap just in case.
-        let origin_stdin_flags = origin_stdin_flags?;
-
-        termios::tcsetattr(
-            STDIN_FILENO,
-            termios::SetArg::TCSAFLUSH,
-            &origin_stdin_flags,
-        )?;
-    }
-
-    session.set_echo(origin_pty_echo)?;
-
-    result
-}
-
-// copy paste of sync version with async await syntax
-#[cfg(all(unix, feature = "async"))]
-async fn interact<R, W>(
-    session: &mut Session,
-    mut options: InteractOptions<R, W>,
-) -> Result<WaitStatus, Error>
-where
-    R: futures_lite::AsyncRead + std::marker::Unpin,
-    W: Write,
-{
-    use futures_lite::{AsyncReadExt, AsyncWriteExt};
-
-    let options_has_input_checks = !options.input_handlers.is_empty();
-    let mut input_buffer = if options_has_input_checks {
-        Some(Vec::new())
-    } else {
-        None
-    };
-
-    let mut output_buffer = Vec::new();
-
-    let mut buf = [0; 512];
-    loop {
-        let status = session.status()?;
-        if !matches!(status, WaitStatus::StillAlive) {
-            return Ok(status);
-        }
-
-        match session.try_read(&mut buf).await {
-            Ok(n) => {
-                let eof = n == 0;
-
-                output_buffer.extend_from_slice(&buf[..n]);
-                options.check_output(session, &mut output_buffer, eof)?;
-
-                if n == 0 {
-                    return Ok(status);
-                }
-
-                let bytes = if let Some(filter) = options.output_filter.as_mut() {
-                    (filter)(&buf[..n])?
-                } else {
-                    Cow::Borrowed(&buf[..n])
-                };
-
-                options.output.write_all(&bytes)?;
-                options.output.flush()?;
-            }
-            Err(err) if err.kind() == io::ErrorKind::WouldBlock => {}
-            Err(err) => return Err(err.into()),
-        }
-
-        // We dont't print user input back to the screen.
-        // In terminal mode it will be ECHOed back automatically.
-        // This way we preserve terminal seetings for example when user inputs password.
-        // The terminal must have been prepared before.
-        match options.input.read(&mut buf).await {
-            Ok(0) => {
-                return Ok(status);
-            }
-            Ok(n) => {
-                let bytes = &buf[..n];
-                let bytes = if let Some(filter) = options.input_filter.as_mut() {
-                    (filter)(bytes)?
-                } else {
-                    Cow::Borrowed(bytes)
-                };
-
-                let buffer = if let Some(check_buffer) = input_buffer.as_mut() {
-                    check_buffer.extend_from_slice(&bytes);
-                    loop {
-                        match options.check_input(session, check_buffer)? {
-                            Match::Yes(n) => {
-                                check_buffer.drain(..n);
-                                if check_buffer.is_empty() {
-                                    break vec![];
-                                }
-                            }
-                            Match::No => {
-                                let buffer = check_buffer.to_vec();
-                                check_buffer.clear();
-                                break buffer;
-                            }
-                            Match::MaybeLater => break vec![],
-                        }
-                    }
-                } else {
-                    bytes.to_vec()
-                };
-
-                let escape_char_position =
-                    buffer.iter().position(|c| *c == options.escape_character);
-                match escape_char_position {
-                    Some(pos) => {
-                        session.write_all(&buffer[..pos]).await?;
-                        return Ok(status);
-                    }
-                    None => {
-                        session.write_all(&buffer[..]).await?;
-                    }
-                }
-            }
-            Err(err) if err.kind() == io::ErrorKind::WouldBlock => {}
-            Err(err) => return Err(err.into()),
-        }
-
-        options.call_idle_handler(session)?;
-    }
-}
-
-#[cfg(windows)]
-fn interact_in_terminal<R, W, C>(
-    session: &mut Session,
-    options: InteractOptions<R, W, C>,
-) -> Result<(), Error>
-where
-    R: Read,
-    W: Write,
-{
-    // flush buffers
-    session.flush()?;
-
-    let console = conpty::console::Console::current()?;
-    console.set_raw()?;
-
-    let r = interact(session, options);
-
-    console.reset()?;
-
-    r
-}
-
-// copy paste of unix version with changed return type
-#[cfg(windows)]
-fn interact<R, W, C>(
-    session: &mut Session,
-    mut options: InteractOptions<R, W, C>,
-) -> Result<(), Error>
-where
-    R: Read,
-    W: Write,
-{
-    let options_has_input_checks = !options.input_handlers.is_empty();
-    let mut input_buffer = if options_has_input_checks {
-        Some(Vec::new())
-    } else {
-        None
-    };
-
-    let mut output_buffer = Vec::new();
-
-    let mut buf = [0; 512];
-    loop {
-        match session.try_read(&mut buf) {
-            Ok(n) => {
-                let eof = n == 0;
-
-                output_buffer.extend_from_slice(&buf[..n]);
-                options.check_output(session, &mut output_buffer, eof)?;
-
-                if n == 0 {
-                    return Ok(());
-                }
-
-                let bytes = if let Some(filter) = options.output_filter.as_mut() {
-                    (filter)(&buf[..n])?
-                } else {
-                    Cow::Borrowed(&buf[..n])
-                };
-
-                options.output.write_all(&bytes)?;
-                options.output.flush()?;
-            }
-            Err(err) if err.kind() == io::ErrorKind::WouldBlock => {}
-            Err(err) => return Err(err.into()),
-        }
-
-        // We dont't print user input back to the screen.
-        // In terminal mode it will be ECHOed back automatically.
-        // This way we preserve terminal seetings for example when user inputs password.
-        // The terminal must have been prepared before.
-        match options.input.read(&mut buf) {
+        match input.read(&mut buf) {
             Ok(0) => {
                 return Ok(());
             }
@@ -747,9 +425,9 @@ where
                 let buffer = if let Some(check_buffer) = input_buffer.as_mut() {
                     check_buffer.extend_from_slice(&bytes);
                     loop {
-                        match options.check_input(session, check_buffer)? {
+                        match options.check_input(input, output, session, check_buffer)? {
                             Match::Yes(n) => {
-                                check_buffer.drain(..n);
+                                let _ = check_buffer.drain(..n);
                                 if check_buffer.is_empty() {
                                     break vec![];
                                 }
@@ -782,86 +460,125 @@ where
             Err(err) => return Err(err.into()),
         }
 
-        options.call_idle_handler(session)?;
+        options.call_idle_handler(input, output, session)?;
     }
 }
 
-/// A non blocking version of STDIN.
-///
-/// It's not recomended to be used directly.
-/// But we expose it because its used in [InteractOptions::terminal]
-#[cfg(unix)]
-pub struct NonBlockingStdin {
-    stream: Stream,
-}
-
-#[cfg(unix)]
-impl NonBlockingStdin {
-    fn new() -> Result<Self, Error> {
-        // it's crusial to make a DUP call here.
-        // If we don't actual stdin will be closed,
-        // And any interaction with it may cause errors.
-        //
-        // Why we don't use a `std::fs::File::try_clone` with a 0 fd?
-        // Because for some reason it actually doesn't make the same things as DUP does,
-        // eventhough a research showed that it should.
-        // https://github.com/zhiburt/expectrl/issues/7#issuecomment-884787229
-        let stdin_copy_fd = dup(STDIN_FILENO)?;
-        let stdin = unsafe { std::fs::File::from_raw_fd(stdin_copy_fd) };
-        let stream = Stream::new(stdin);
-
-        Ok(Self { stream })
-    }
-}
-
-#[cfg(unix)]
-#[cfg(not(feature = "async"))]
-impl Read for NonBlockingStdin {
-    fn read(&mut self, buf: &mut [u8]) -> io::Result<usize> {
-        self.stream.try_read(buf)
-    }
-}
-
-#[cfg(unix)]
+// copy paste of sync version with async await syntax
 #[cfg(feature = "async")]
-impl futures_lite::AsyncRead for NonBlockingStdin {
-    fn poll_read(
-        mut self: std::pin::Pin<&mut Self>,
-        cx: &mut std::task::Context<'_>,
-        buf: &mut [u8],
-    ) -> std::task::Poll<io::Result<usize>> {
-        use futures_lite::FutureExt;
-        Box::pin(self.stream.try_read(buf)).poll(cx)
-    }
-}
+async fn interact<P, S, R, W, C>(
+    options: &mut InteractOptions<P, S, R, W, C>,
+    session: &mut Session<P, S>,
+    input: &mut R,
+    output: &mut W,
+) -> Result<(), Error>
+where
+    P: Healthcheck + Unpin,
+    S: futures_lite::AsyncRead + futures_lite::AsyncWrite + Unpin,
+    R: futures_lite::AsyncRead + Unpin,
+    W: Write,
+{
+    use futures_lite::{AsyncReadExt, AsyncWriteExt};
 
-/// See unix version
-#[cfg(windows)]
-pub struct NonBlockingStdin {
-    current_terminal: Console,
-}
+    let mut output_buffer = Vec::new();
+    let options_has_input_checks = !options.input_handlers.is_empty();
+    let mut input_buffer = if options_has_input_checks {
+        Some(Vec::new())
+    } else {
+        None
+    };
+    let mut exited = false;
 
-#[cfg(windows)]
-impl NonBlockingStdin {
-    fn new() -> Result<Self, Error> {
-        let console = conpty::console::Console::current()?;
-        Ok(Self {
-            current_terminal: console,
-        })
-    }
-}
-
-#[cfg(windows)]
-impl Read for NonBlockingStdin {
-    fn read(&mut self, buf: &mut [u8]) -> io::Result<usize> {
-        // we can't easily read in non-blocking manner,
-        // but we can check when there's something to read,
-        // which seems to be enough to not block.
-        if self.current_terminal.is_stdin_not_empty()? {
-            io::stdin().read(buf)
-        } else {
-            Err(io::Error::new(io::ErrorKind::WouldBlock, ""))
+    let mut buf = [0; 512];
+    loop {
+        // In case where proceses exits we are trying to
+        // fill buffer to run callbacks if there was something in.
+        //
+        // We ignore errors because there might be errors like EOCHILD etc.
+        let status = session.is_alive();
+        if matches!(status, Ok(false)) {
+            exited = true;
         }
+
+        if let Some(result) = futures_lite::future::poll_once(session.read(&mut buf)).await {
+            let n = result?;
+            let eof = n == 0;
+            if eof {
+                exited = true;
+            }
+
+            output_buffer.extend_from_slice(&buf[..n]);
+            options.check_output(input, output, session, &mut output_buffer, eof)?;
+
+            let bytes = if let Some(filter) = options.output_filter.as_mut() {
+                (filter)(&buf[..n])?
+            } else {
+                Cow::Borrowed(&buf[..n])
+            };
+
+            output.write_all(&bytes)?;
+            output.flush()?;
+        }
+
+        if exited {
+            return Ok(());
+        }
+
+        // We dont't print user input back to the screen.
+        // In terminal mode it will be ECHOed back automatically.
+        // This way we preserve terminal seetings for example when user inputs password.
+        // The terminal must have been prepared before.
+        match input.read(&mut buf).await {
+            Ok(0) => {
+                return Ok(());
+            }
+            Ok(n) => {
+                let bytes = &buf[..n];
+                let bytes = if let Some(filter) = options.input_filter.as_mut() {
+                    (filter)(bytes)?
+                } else {
+                    Cow::Borrowed(bytes)
+                };
+
+                let buffer = if let Some(check_buffer) = input_buffer.as_mut() {
+                    check_buffer.extend_from_slice(&bytes);
+                    loop {
+                        match options.check_input(input, output, session, check_buffer)? {
+                            Match::Yes(n) => {
+                                check_buffer.drain(..n);
+                                if check_buffer.is_empty() {
+                                    break vec![];
+                                }
+                            }
+                            Match::No => {
+                                let buffer = check_buffer.to_vec();
+                                check_buffer.clear();
+                                break buffer;
+                            }
+                            Match::MaybeLater => break vec![],
+                        }
+                    }
+                } else {
+                    bytes.to_vec()
+                };
+
+                let escape_char_position =
+                    buffer.iter().position(|c| *c == options.escape_character);
+                match escape_char_position {
+                    Some(pos) => {
+                        session.write_all(&buffer[..pos]).await?;
+                        return Ok(());
+                    }
+                    None => {
+                        session.write_all(&buffer[..]).await?;
+                    }
+                }
+            }
+            Err(err) if err.kind() == io::ErrorKind::WouldBlock => {}
+            Err(err) => return Err(err.into()),
+        }
+
+        options.call_idle_handler(input, output, session)?;
     }
 }
 
