@@ -1,4 +1,4 @@
-use expectrl::spawn;
+use expectrl::{session::Session, spawn};
 use std::{thread, time::Duration};
 
 #[cfg(feature = "async")]
@@ -51,20 +51,19 @@ fn send() {
 #[test]
 fn send() {
     let mut session = spawn("powershell -C type").unwrap();
-    session.write(b"Hello World").unwrap();
-
-    thread::sleep(Duration::from_millis(300));
-
-    let mut buf = vec![0; 1028];
-    let _ = session.read(&mut buf).unwrap();
-    let n = session.read(&mut buf).unwrap();
-
-    let s = String::from_utf8_lossy(&buf[..n]);
-    if !s.contains("Hello World") {
-        panic!(
-            "Expected to get {:?} in the output, but got {:?}",
-            "Hello World", s
-        );
+    #[cfg(not(feature = "async"))]
+    {
+        session.write(b"Hello World").unwrap();
+        thread::sleep(Duration::from_millis(300));
+        session.expect("Hello World").unwrap();
+    }
+    #[cfg(feature = "async")]
+    {
+        futures_lite::future::block_on(async {
+            session.write(b"Hello World").await.unwrap();
+            thread::sleep(Duration::from_millis(300));
+            session.expect("Hello World").await.unwrap();
+        })
     }
 }
 
@@ -112,17 +111,39 @@ fn send_multiline() {
 #[test]
 fn send_multiline() {
     let mut session = spawn("powershell -C type").unwrap();
-    session.send("Hello World\r\n").unwrap();
+    #[cfg(not(feature = "async"))]
+    {
+        session.send("Hello World\r\n").unwrap();
 
-    thread::sleep(Duration::from_millis(300));
+        thread::sleep(Duration::from_millis(300));
 
-    let buf = session.lines().nth(2).unwrap().unwrap();
+        let buf = session.lines().nth(2).unwrap().unwrap();
 
-    if !buf.contains("Hello World") {
-        panic!(
-            "Expected to get {:?} in the output, but got {:?}",
-            "Hello World", buf
-        );
+        if !buf.contains("Hello World") {
+            panic!(
+                "Expected to get {:?} in the output, but got {:?}",
+                "Hello World", buf
+            );
+        }
+    }
+    #[cfg(feature = "async")]
+    {
+        use futures_lite::{AsyncBufReadExt, StreamExt};
+
+        futures_lite::future::block_on(async {
+            session.send("Hello World\r\n").await.unwrap();
+
+            thread::sleep(Duration::from_millis(300));
+
+            let buf = session.lines().nth(2).await.unwrap().unwrap();
+
+            if !buf.contains("Hello World") {
+                panic!(
+                    "Expected to get {:?} in the output, but got {:?}",
+                    "Hello World", buf
+                );
+            }
+        })
     }
 }
 
@@ -176,16 +197,76 @@ fn send_line() {
 #[test]
 fn send_line() {
     let mut session = spawn("powershell -C type").unwrap();
-    session.send_line("Hello World").unwrap();
+    #[cfg(not(feature = "async"))]
+    {
+        session.send_line("Hello World").unwrap();
 
-    thread::sleep(Duration::from_millis(300));
+        thread::sleep(Duration::from_millis(300));
 
-    let buf = session.lines().nth(2).unwrap().unwrap();
+        let buf = session.lines().nth(2).unwrap().unwrap();
 
-    if !buf.contains("Hello World") {
-        panic!(
-            "Expected to get {:?} in the output, but got {:?}",
-            "Hello World", buf
-        );
+        if !buf.contains("Hello World") {
+            panic!(
+                "Expected to get {:?} in the output, but got {:?}",
+                "Hello World", buf
+            );
+        }
+    }
+    #[cfg(feature = "async")]
+    {
+        use futures_lite::{AsyncBufReadExt, StreamExt};
+
+        futures_lite::future::block_on(async {
+            session.send_line("Hello World").await.unwrap();
+
+            thread::sleep(Duration::from_millis(300));
+
+            let buf = session.lines().nth(2).await.unwrap().unwrap();
+
+            if !buf.contains("Hello World") {
+                panic!(
+                    "Expected to get {:?} in the output, but got {:?}",
+                    "Hello World", buf
+                );
+            }
+        })
+    }
+}
+
+#[test]
+fn test_spawn_no_command() {
+    #[cfg(unix)]
+    assert!(spawn("").is_err());
+    #[cfg(windows)]
+    assert!(spawn("").is_ok());
+}
+
+#[test]
+#[ignore = "it's a compile time check"]
+fn test_session_as_writer() {
+    #[cfg(not(feature = "async"))]
+    {
+        let _: Box<dyn std::io::Write> = Box::new(spawn("ls").unwrap());
+        let _: Box<dyn std::io::Read> = Box::new(spawn("ls").unwrap());
+        let _: Box<dyn std::io::BufRead> = Box::new(spawn("ls").unwrap());
+
+        fn _io_copy(mut session: Session) {
+            let _ = std::io::copy(&mut std::io::empty(), &mut session).unwrap();
+        }
+    }
+    #[cfg(feature = "async")]
+    {
+        let _: Box<dyn futures_lite::AsyncWrite> =
+            Box::new(spawn("ls").unwrap()) as Box<dyn futures_lite::AsyncWrite>;
+        let _: Box<dyn futures_lite::AsyncRead> =
+            Box::new(spawn("ls").unwrap()) as Box<dyn futures_lite::AsyncRead>;
+        let _: Box<dyn futures_lite::AsyncBufRead> =
+            Box::new(spawn("ls").unwrap()) as Box<dyn futures_lite::AsyncBufRead>;
+
+        async fn _io_copy(mut session: Session) {
+            futures_lite::io::copy(futures_lite::io::empty(), &mut session)
+                .await
+                .unwrap();
+        }
     }
 }
